from collections import defaultdict
from typing import Dict, Set, List

from variable_protocols.protocols import Variable, struct_check, fmt
from variable_protocols.transformations import Transformation, Tr

LookUpTable = Dict[Variable, Set[Transformation]]
lookup_table_source: LookUpTable = defaultdict(set)
lookup_table_target: LookUpTable = defaultdict(set)


def register_(transformation: Transformation):
    lookup_table_source[transformation.source].add(transformation)
    lookup_table_target[transformation.source].add(transformation)


def look_up(source: Variable, target: Variable, max_depth: int = 16) -> List[Tr]:
    if struct_check(source, target):
        raise ValueError(f"Source and target are the same: {fmt(source)}")
    else:
        reachable_s_: Dict[Variable, List[Set[Tr]]] = {}
        reachable_t_: Dict[Variable, List[Set[Tr]]] = {}
        depth_ = 0
        while depth_ < max_depth:
            for i, r in reachable_s_.items():
                tr = lookup_table_source[i]
                reachable_s_[i] = r + [tr]
            common = set(reachable_s_).intersection(set(reachable_t_))
            if len(common) >= 1:
                tr_list: List[List[Set[Tr]]] = [reachable_s_[k] + reachable_t_[k]
                                                for k in common]
                break
        else:
            raise Exception("Max-depth is reached when looking up for transformations")

        if len(tr_list) == 0:
            raise Exception(f"No transformation are found from {source} to {target}")
        elif len(tr_list) == 1:
            tr_series = tr_list[0]
            if all(len(tr_set) == 1 for tr_set in tr_series):
                return [next(iter(trs)) for trs in tr_series]
            else:
                msg = ""
                for trs in tr_series:
                    nxt = next(iter(trs))
                    msg += f"Transformations that are found from {nxt.source} to {nxt.target}:\n"
                    msg += f"  {trs}"
                raise Exception(msg)
        else:
            raise Exception("Multiple transformations have been retrieved: \n"
                            + "".join(f"  {tr}\n" for tr in tr_list))



from __future__ import annotations

from typing import List, Dict, Set, Literal, Iterable, Optional

from variable_protocols.base_variables import BaseVariable, \
    OneSideSupported, Gamma, Bounded, OneHot, NamedCategorical, \
    CategoryIds, Ordinal, CategoricalVector, Gaussian
from variable_protocols.protocols import VariableTensor, Variable, \
    VariableGroup, DimensionFamily


def bounded_float(min_val: float, max_val: float) -> Bounded:
    if min_val > max_val:
        raise ValueError(f"min_val > max_val: {min_val} > {max_val}")
    return Bounded(max_val, min_val)


def one_side_supported(bound: int, max_or_min: Literal["max", "min"]) -> OneSideSupported:
    return OneSideSupported(bound, max_or_min)


def positive_float() -> OneSideSupported:
    return OneSideSupported(0, "min")


def negative_float() -> OneSideSupported:
    return OneSideSupported(0, "max")


def gamma(alpha: float, beta: float) -> Gamma:
    if alpha > 0 and beta > 0:
        return Gamma(alpha, beta)
    else:
        raise ValueError(f"Invalid alpha or beta value {alpha}, {beta}")


def erlang(k: int, beta: float) -> Gamma:
    if k > 0:
        return gamma(k, beta)
    else:
        raise ValueError(f"Invalid k:{k}")


def exponential(lambda_: float) -> Gamma:
    if lambda_ > 0:
        return erlang(1, lambda_)
    else:
        raise ValueError(f"Invalid lambda(l) value {lambda_}")


def ordinal(n_category: int) -> Ordinal:
    if n_category > 0:
        return Ordinal(n_category)
    else:
        raise ValueError(f"Invalid n_category value: {n_category}")


def one_hot(n_category: int) -> OneHot:
    if n_category > 0:
        return OneHot(n_category)
    else:
        raise ValueError(f"Invalid n_category value: {n_category}")


def cat_vec(n_category: int, n_embedding: int) -> CategoricalVector:
    if n_category > 0:
        if n_embedding > 0:
            return CategoricalVector(n_category, n_embedding)
        else:
            raise ValueError(f"Invalid n_embedding value: {n_embedding}")
    else:
        raise ValueError(f"Invalid n_category value: {n_category}")


def cat_from_names(names: Iterable[str]) -> NamedCategorical:
    return NamedCategorical(frozenset(names))


def cat_ids(max_id_len: int) -> CategoryIds:
    return CategoryIds(max_id_len)


def gaussian(mean: float, var: float) -> Gaussian:
    if var > 0:
        return Gaussian(mean, var)
    else:
        raise ValueError(f"Invalid variance(var) value: {var}")


def dim(name: str,
        length: Optional[int],
        positioned: bool = True,
        n_members: int = 1) -> DimensionFamily:
    return DimensionFamily(label=name, len=length,
                           n_members=n_members, positioned=positioned)


def var_tensor(var: BaseVariable, dims: Set[DimensionFamily]) -> VariableTensor:
    return VariableTensor(var, frozenset(dims))


def var_scalar(var: BaseVariable) -> VariableTensor:
    return VariableTensor(var, frozenset())


def var_group(vars_set: Set[Variable]) -> VariableGroup:
    assert isinstance(vars_set, set)
    if len(vars_set) <= 1:
        raise ValueError("A variable set/group must contain more than one variable")
    return VariableGroup(frozenset(vars_set))


def var_unique(var: Variable, name: str) -> VariableTensor:
    return VariableTensor(var, frozenset(), label=name)


def var_dict(vars_dict: Dict[Variable, str]) -> VariableGroup:
    return var_group({
        var_unique(var, name)
        for var, name in vars_dict.items()
    })


def var_ordered(vars_list: List[Variable]) -> VariableGroup:
    if not len(vars_list) == len(set(vars_list)):
        raise ValueError("There are duplicate variables in input,"
                         "use var_tensor to combine them first,"
                         "or var_unique to make them unique")
    return var_group({
        var_unique(var, str(i))
        for i, var in enumerate(vars_list)
    })



from __future__ import annotations

from collections import defaultdict
from typing import List, Tuple, Optional, FrozenSet, Any, Dict, Set

from variable_protocols.protocols import Variable, VariableGroup, struct_hash, VariableTensor


# noinspection PyTypeChecker
# because pycharm sucks
class HashedTree:
    def __init__(self, var: Variable,
                 ignore_names: bool = True,
                 is_root: bool = True) -> None:
        self.hash = struct_hash(var, ignore_names)
        self.is_root = is_root
        self.id: Optional[int] = None

        if var.type == 'VariableTensor':
            assert isinstance(var, VariableTensor)
            if var.var.type == 'BaseVariable':
                self.children: List[HashedTree] = []
                self.nodes: List[HashedTree] = [self]
            else:
                assert isinstance(var.var, VariableGroup)
                subtree = HashedTree(var.var)
                self.children = [subtree]
                self.nodes = self.children + subtree.nodes

        elif var.type == 'VariableGroup':
            assert isinstance(var, VariableGroup)
            self.children = list(sorted(
                (HashedTree(var, ignore_names, is_root=False) for var in var.vars),
                key=lambda x: x.hash
            ))
            self.nodes = sum([c.nodes for c in self.children], [])
        else:
            raise Exception(f"Unexpected Variable type {var.type}")

        if is_root:
            self.node_idx: Optional[Dict[HashedTree, int]] = {node: i for i, node in enumerate(self.nodes)}
            self.hash_dict: Optional[Dict[str, Set[HashedTree]]] = defaultdict(set)
            for n in self.nodes:
                self.hash_dict[n.hash].add(n)
        else:
            self.hash_dict = None
            self.node_idx = None


def check_tr(t1: HashedTree, t2: HashedTree) -> bool:
    return t1.hash == t2.hash


DiffNode = HashedTree


def diff(t1: HashedTree, t2: HashedTree) -> Tuple[Dict[str, int], Dict[str, int]]:
    if t1.hash_dict is None or t2.hash_dict is None:
        raise ValueError("Either tree entered for compare is not root")
    else:
        d1, d2 = t1.hash_dict, t2.hash_dict
        common_keys = set(d1).intersection(set(d2))
        t1_special = {k: len(d1[k]) for k in set(d1) - common_keys}
        t2_special = {k: len(d2[k]) for k in set(d2) - common_keys}
        for k in common_keys:
            if len(d1[k]) > len(d2[k]):
                t1_special[k] = len(d1[k]) - len(d2[k])
            elif len(d1[k]) < len(d2[k]):
                t2_special[k] = len(d2[k]) - len(d1[k])
        return t1_special, t2_special


DiffResultTr = List[Tuple[FrozenSet[DiffNode], FrozenSet[DiffNode]]]


def diff_tr(t1: HashedTree, t2: HashedTree) -> DiffResultTr:
    assert t1.is_root
    assert t2.is_root
    return diff_helper(t1, t2)


def diff_helper(t1: HashedTree, t2: HashedTree) -> DiffResultTr:
    if t1.hash == t2.hash:
        return []
    elif t1.children is None:
        if t2.children is not None:
            return [(frozenset([t1]), frozenset([t2]))]
        return []
    elif t2.children is None:
        return [(frozenset([t1]), frozenset([t2]))]
    else:
        tc1 = set(t1.children)
        tc2 = set(t2.children)
        tc1_hd = {n.hash: n for n in t1.children}
        tc2_hd = {n.hash: n for n in t2.children}
        g1, g2 = set(), set()
        for n1 in t1.children:
            if n1.hash in tc2_hd:
                print(n1.hash, tc2_hd[n1.hash].hash)
                print(n1, tc2_hd[n1.hash])
                tc2.remove(tc2_hd[n1.hash])
                tc1.remove(n1)
            else:
                g1.add(n1)
        for n2 in t2.children:
            if n2.hash in tc1_hd:
                tc1.remove(tc1_hd[n2.hash])
                tc2.remove(n2)
            else:
                g2.add(n2)
        if len(tc1) == 0 and len(tc2) == 0:
            raise Exception("This should not happen")
        return [(frozenset(tc1), frozenset(tc2))]



from variable_protocols.protocols import Variable
from variable_protocols.variables import var_tensor, dim, bounded_float, ordinal, one_hot, \
    cat_ids, cat_vec


# noinspection PyTypeChecker
# because pycharm sucks
def var_image(h: int, w: int, n_channels: int = 3, int_format: bool = False) -> Variable:
    if int_format:
        return var_tensor(bounded_float(0, 1), {dim("h", h), dim("w", w), dim("channel", n_channels)})
    else:
        return var_tensor(ordinal(256), {dim("h", h), dim("w", w), dim("channel", n_channels)})


# noinspection PyTypeChecker
# because pycharm sucks
def var_sentence_1hot(n_vocab: int) -> Variable:
    return var_tensor(one_hot(n_vocab), {dim("word", None)})


# noinspection PyTypeChecker
# because pycharm sucks
def var_sentence(longest_word_len: int) -> Variable:
    return var_tensor(cat_ids(longest_word_len), {dim("word", None)})


# noinspection PyTypeChecker
# because pycharm sucks
def var_sentence_embeddings(n_vocab: int, size_embedding: int) -> Variable:
    return var_tensor(cat_vec(n_vocab, size_embedding), {dim("word", None)})



import abc
from typing import Literal, Protocol, NamedTuple, FrozenSet, runtime_checkable

BaseVariableType = Literal['BaseVariable']


@runtime_checkable
class BaseVariable(Protocol):
    @property
    @abc.abstractmethod
    def type_name(self) -> str: ...

    @property
    @abc.abstractmethod
    def type(self) -> BaseVariableType: ...

    @abc.abstractmethod
    def fmt(self) -> str: ...

    @abc.abstractmethod
    def _asdict(self) -> dict: ...


@runtime_checkable
class CustomHashBaseVariable(Protocol):
    @property
    @abc.abstractmethod
    def type_name(self) -> str: ...

    @property
    @abc.abstractmethod
    def type(self) -> BaseVariableType: ...

    @abc.abstractmethod
    def fmt(self) -> str: ...

    @abc.abstractmethod
    def struct_hash(self, ignore_names: bool) -> str: ...


class OneSideSupported(NamedTuple):
    bound: float
    min_or_max: Literal["min", "max"]
    type_name: str = 'one_side_supported'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        if self.min_or_max == 'min':
            return f"OneSideSupported(>={self.bound})"
        else:
            assert self.min_or_max == 'max'
            return f"OneSideSupported(<={self.bound})"


class Gamma(NamedTuple):
    alpha: float
    beta: float
    type_name: str = 'gamma'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"Gamma({self.alpha}, {self.beta})"


class Bounded(NamedTuple):
    max: float
    min: float
    type_name: str = 'bounded'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"Bounded({self.min}, {self.max})"


class OneHot(NamedTuple):
    n_category: int
    type_name: str = '1hot'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"OneHot({self.n_category})"


class NamedCategorical(NamedTuple):
    names: FrozenSet[str]
    type_name: str = 'named_categorical'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"Categorical({', '.join(self.names)})"

    def struct_hash(self, ignore_names: bool) -> str:
        content = (
            str(len(self.names)) if ignore_names
            else f"[{','.join(self.names)}]")
        return f"B[{self.type_name}|{content}]"


class CategoryIds(NamedTuple):
    max_id_len: int
    type_name: str = 'category_ids'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"CategoryIds(max_id_len={self.max_id_len})"


class Ordinal(NamedTuple):
    n_category: int
    type_name: str = 'ordinal'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"Ordinal({self.n_category})"


class CategoricalVector(NamedTuple):
    n_category: int
    n_embedding: int
    type_name: str = '2vec'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"CategoricalVector({self.n_category}, n_embedding={self.n_embedding})"


class Gaussian(NamedTuple):
    mean: float = 0
    var: float = 1
    type_name: str = 'gaussian'
    type: Literal['BaseVariable'] = 'BaseVariable'

    def fmt(self) -> str:
        return f"Gaussian({self.mean}, {self.var})"


def struct_hash_base_variable(var: BaseVariable, ignore_names: bool) -> str:
    if isinstance(var, CustomHashBaseVariable):
        return var.struct_hash(ignore_names)
    else:
        # noinspection PyProtectedMember
        # Because Pycharm sucks
        var_dict = var._asdict()
        var_dict.pop("type")
        var_type = var_dict.pop("type_name")
        content = "|".join(str(int(v)) if isinstance(v, bool) else str(v)
                           for _, v in var_dict.items())
        return f"B[{var_type}|{content}]"



from __future__ import annotations
from abc import abstractmethod
from typing import Protocol, List, Optional, NamedTuple

from variable_protocols.protocols import Variable, struct_check, fmt


class Transformation(Protocol):
    @property
    @abstractmethod
    def source(self) -> Variable: ...

    @property
    @abstractmethod
    def target(self) -> Variable: ...

    @property
    @abstractmethod
    def name(self) -> str: ...


Tr = Transformation


class NewTransformation(NamedTuple):
    source: Variable
    target: Variable
    name: str


def new_transformation(source: Variable, target: Variable, name: str) -> Tr:
    # noinspection PyTypeChecker
    # cus Pycharm sucks
    return NewTransformation(source, target, name)


def check(transformation: Transformation,
          source: Variable,
          target: Variable,
          ignore_names: bool = False) -> bool:
    return struct_check(transformation.source, source, ignore_names) and \
           struct_check(transformation.target, target, ignore_names)


def transform(transformation: Transformation,
              source: Variable) -> Variable:
    if not struct_check(transformation.source, source):
        raise ValueError(f"Expected source of \"{fmt(transformation.source)}\","
                         f" got \"{fmt(source)}\"")
    return transformation.target


def pipe(trs: List[Transformation], name: Optional[str] = None) -> Tr:
    wrong_list = []
    for back, front in zip(trs[:-1], trs[:-1]):
        if not back.target == front.source:
            wrong_list.append((back.target, front.source))

    if len(wrong_list) > 0:
        mismatch_msgs = "".join(f"\n  {v[0]} != {v[1]}" for v in wrong_list)
        raise ValueError(f"Unable to make pipe. "
                         f"Found {len(wrong_list)} mismatching Transformations: {mismatch_msgs}")
    else:
        name = "".join(t.name for t in trs) if not name else name
        return new_transformation(
            source=trs[0].source,
            target=trs[-1].target,
            name=name
        )



from __future__ import annotations
import abc
from typing import Literal, Protocol, NamedTuple, FrozenSet, Optional
from variable_protocols.base_variables import BaseVariable, struct_hash_base_variable

VariableType = Literal['VariableGroup',
                       'VariableTensor']

TensorBaseType = Literal['BaseVariable',
                         'VariableGroup',
                         'VariableTensor']


class TensorBase(Protocol):
    @property
    @abc.abstractmethod
    def type(self) -> TensorBaseType: ...


class Variable(Protocol):
    @property
    @abc.abstractmethod
    def id(self) -> Optional[str]: ...

    @property
    @abc.abstractmethod
    def label(self) -> Optional[str]: ...

    @property
    @abc.abstractmethod
    def type(self) -> VariableType: ...

    @abc.abstractmethod
    def __hash__(self) -> int: ...


# Dimension Group
class DimensionFamily(NamedTuple):
    positioned: bool
    label: str
    len: Optional[int]
    n_members: int = 1

    @property
    def struct_hash(self) -> str:
        return f"D[{self.len}|{self.n_members}|{int(self.positioned)}]"

    @property
    def id(self) -> str:
        return str(hash(self))[:5]

    def fmt(self) -> str:
        family = f"*{self.n_members}" if self.n_members > 1 else ""
        positioned = "|shuffle-able" if not self.positioned else ""
        return f"{self.label}[{self.len}{positioned}]{family}"


# Grouping Variables of the same type
class VariableTensor(NamedTuple):
    var: TensorBase
    dims: FrozenSet[DimensionFamily]
    label: Optional[str] = None
    type: Literal['VariableTensor'] = 'VariableTensor'

    @property
    def id(self) -> str:
        return str(hash(self))[-4:]

    @property
    def struct_hash(self) -> str:
        if len(self.dims) == 0:
            return f"T[{struct_hash(self.var)}]"
        else:
            dims = ",".join([d.struct_hash for d in self.dims])
            return f"T[{struct_hash(self.var)}|[{dims}]]"

    def fmt(self, indent: int = 2, curr_indent: int = 0) -> str:
        var_type = fmt(self.var)
        if len(self.dims) == 0:
            if self.label is None:
                return curr_indent * " " + var_type
            else:
                return curr_indent * " " + f"{self.label}: {var_type}"
        else:
            header = "Tensor#"
            dims = ", ".join(d.fmt() for d in self.dims)
            if len(header) + len(var_type) + len(": ") + len(dims) > 70:
                indent_spaces = (curr_indent + len(header) + indent) * ' '
                return f"{header}{var_type}:\n{indent_spaces}{dims}"
            else:
                return f"{header}{var_type}: {dims}"

    def __hash__(self) -> int:
        # noinspection PyTypeChecker
        # because pyCharm sucks
        return hash(struct_hash(self))


# Grouping Variables of different types
class VariableGroup(NamedTuple):
    vars: FrozenSet[Variable]
    label: Optional[str] = None
    type: Literal['VariableGroup'] = 'VariableGroup'

    @property
    def id(self) -> str:
        return str(hash(self))[-4:]

    @property
    def struct_hash(self) -> str:
        _vars = ",".join([struct_hash(v) for v in self.vars])
        return f"G[{_vars}]"

    @classmethod
    def make(cls,
             _vars: FrozenSet[Variable],
             label: Optional[str] = None) -> VariableGroup:
        if len(_vars) > 1:
            return VariableGroup(_vars, label)
        else:
            raise ValueError("Groups must be made of two Variables")

    def fmt(self, indent: int = 2, curr_indent: int = 0) -> str:
        s = "Set{\n"
        curr_indent += indent
        for i, var in enumerate(self.vars):
            if self.label is None:
                s += f"{fmt(var, indent=indent, curr_indent=curr_indent)},\n"
            else:
                s += f"{self.label}: {fmt(var, indent=indent, curr_indent=curr_indent)},\n"
        s = s.strip(",\n")
        s += "}\n"
        return s

    def __hash__(self) -> int:
        # noinspection PyTypeChecker
        # because pyCharm sucks
        return hash(struct_hash(self))


def struct_hash(var: TensorBase, ignore_names: bool = False) -> str:
    if var.type == 'BaseVariable':
        assert isinstance(var, BaseVariable)
        return struct_hash_base_variable(var, ignore_names)
    elif var.type == 'VariableTensor':
        assert isinstance(var, VariableTensor)
        return var.struct_hash
    else:
        assert isinstance(var, VariableGroup)
        return var.struct_hash


def fmt(var: TensorBase, **kwargs) -> str:
    if var.type == 'BaseVariable':
        assert isinstance(var, BaseVariable)
        return var.fmt()
    elif var.type == 'VariableTensor':
        assert isinstance(var, VariableTensor)
        return var.fmt(**kwargs)
    else:
        assert isinstance(var, VariableGroup)
        return var.fmt(**kwargs)


def struct_check(var1: Variable, var2: Variable, ignore_names: bool = False):
    return struct_hash(var1, ignore_names) == struct_hash(var2, ignore_names)
